/*
 * PS2_keyboard.c
 *
 * Author: JW Technical Consultancy Ltd
 * 
 * Decode PS/2 type keyboard and send to parallel port
 *
 * Important note:  PS2 keyboard command set 2 is implemented
 *                  It may be necessary to use SendCommandToKeyboard() function
 *                  to send enable specific command set for some keyboards
 *                  (I only have one to test)  
 * 
 * 
 */
#include "mcc_generated_files/system.h"
#include "mcc_generated_files/pin_manager.h"
#include "mcc_generated_files/interrupt_manager.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include "global.h"
#include "kbd_international_settings.h"
#include "PS2_keyboard.h"
#include "PS2_KbdProtocol.h"
#include "kbd_parallel_out.h"
#include "Timer.h"


#define LOW_EDGE                    0   /**< Logic zero pulse edge */   
#define HIGH_EDGE                   1   /**< Logic one pulse edge */     

#define DEBUG_KEY_DATA_FRAME    false  /**< If Debug of key data needed */
#define DEBUG_KEY_CODE          false  /**< If output on serial of decoded key code */
#define DEBUG_ASCII_CODE        false  /**< Serial debug ascii key code */

#define DEBUG_UNHANDLED_KEYS    false  /**< Output serial debug infor for unhandled keys */

#define KEY_DATA_SetDigitalOutput()  ( _TRISB11 = 0 )
#define KEY_DATA_SetDigitalInput()   ( _TRISB11 = 1 )
#define KEY_DATA_GetValue()            PORTBbits.RB11
#define KEY_DATA_SetLow()            ( _LATB11 = 0 )
#define KEY_DATA_SetHigh()           ( _LATB11 = 1 )

inline static void KEY_DATA_SetValue(bool value)
{
  if(value)
  {
    KEY_DATA_SetHigh();
  }
  else
  {
    KEY_DATA_SetLow();
  }
}


#define KEY_CLOCK_SetDigitalOutput()  ( _TRISB10 = 0 )
#define KEY_CLOCK_SetDigitalInput()   ( _TRISB10 = 1 )
#define KEY_CLOCK_GetValue()            PORTBbits.RB10
#define KEY_CLOCK_SetLow()            ( _LATB10 = 0 )
#define KEY_CLOCK_SetHigh()           ( _LATB10 = 1 )

inline static void KEY_CLOCK_SetValue(bool value)
{
  if(value)
  {
    KEY_CLOCK_SetHigh();
  }
  else
  {
    KEY_CLOCK_SetLow();
  }
}

#define RX_BITS_EXPECTED          11   /**< Number of bits in PS2 keyboard data RX frame */
#define TX_BITS_EXPECTED          11   /**< Number of bits in PS2 keyboard data TX frame */
#define TX_BITS_TO_SEND           11   /**< Number of bits in PS2 keyboard data TX frame */

#define MAX_FRAMES_BUFFERED        16  /**< Maximum number of key-code data frames buffer */

#define INTER_READ_DELAY_uS       20    /**< Delay after each PS2 frame */
#define REQUEST_TO_SEND_DELAY      2    /**< Time in 100's of uS to hold clock low for ACK to keyboard device */
#define ACK_WAIT_TIMEOUT         200    /**< Time in 100's of uS to wait for acknowledge (20mS, WYSE keyboard acknowledged in 3mS) */
#define CMD_SEND_DELAY           100    /**< Time in 100's of uS to wait before sending next command byte */ 
#define KEYBOARD_ALIVE_INTERVAL  40000  /**< Inverval in 100's of uS between keyboard alive check */

#define FRAME_RX_MASK         0x0001  /**< Mask for bit 0 of frame (start bit) */
#define FRAME_TX_MASK         0x0002  /**< Mask for bit 0 of frame (data 0) */


#define KBD_GET_SCAN_CODE_SET_PARAMETER  0	/**< Get current scan code set */
#define KBD_SCAN_CODE_1	                 1   /**< Set scan code set 1 */
#define KBD_SCAN_CODE_2	                 2   /**< Set scan code set 2 */
#define KBD_SCAN_CODE_3                  3   /**< Set scan code set 3 */

// KBD_READ_ID_CMD // keyboard sends its ID


// Commands sent from keyboard
#define KBD_ERROR                   0x00   /**< Error or output buffer overflow (scan code sets 2 and 3 only) */
#define KBD_ID                      0x83AB /**< Keyboard ID */ 
#define KBD_POWER_UP_TEST_PASSED    0xAA   /**< Power-on Basic Assurance Test Passed */
#define KBD_POWER_UP_TEST_FAILED    0xFC   /**< Power-on Basic Assurance Test Failed */
#define KBD_ECHO                    0xEE   /**< Echo */
#define KBD_RESEND                  0xFE   /**< Resend. Upon receipt of this code, the PS2 Controller will retransmit the previous byte  */
#define KBD_ERROR_BUFFER_OVERFLOW   0xFF   /**< Error or output buffer overflow (scan code set 1 only) */

#define EX_SCAN_CODE_1              0xE0   /**< Extended scan code */
#define EX_SCAN_CODE_2              0xE1   /**< Extended scan code */


#define KEY_BREAK_CMD               0xF0   /**< Scan code received when key released (followed by key scan code)*/
#define KBD_PS2_ACK                 0xFA   /**< Acknowledge received from host command */

#define NOT_IMPLEMENTED             0xFF   /**< Code for key code not implemented */

/************************************************************************************/
/* Key Scan codes                                                                   */
/************************************************************************************/

#define CAPS_LOCK_SCAN_CODE         0x58    /**< Scan code for the caps lock key */
#define NUM_LOCK_SCAN_CODE          0x77    /**< Scan code for the num lock key */
#define SCROLL_LOCK_SCAN_CODE       0x7E    /**< Scan code for the scroll lock key */

#define LEFT_SHIFT_KEY_SCAN_CODE    0x12   /**< Scan code for lfte SHIFT key */
#define RIGHT_SHIFT_KEY_SCAN_CODE   0x59   /**< Scan code for right SHIFT key */
#define LEFT_CTRL_KEY_SCAN_CODE     0x14   /**< Scan code for left CTRL key */
#define RIGHT_CTRL_KEY_SCAN_CODE    0x14   /**< Scan code for right CTRL key (extended code) */
#define LEFT_ALT_KEY_SCAN_CODE      0x11   /**< Scan code for left ALT key */   
#define RIGHT_ALT_KEY_SCAN_CODE     0x11   /**< Scan code for right ALT key */   

// These are extended key codes (0xE0 first) unless on num pad with num lock off

#define BACKSPACE_KEY_SCAN_CODE     0x66    /**< Scan code for the BACK-SPACE key */
#define DELETE_KEY_SCAN_CODE        0x71    /**< Scan code for the NUM-PAD DEL key */
#define EXTENDED_NUM_LOCK           0x12    /**< Scan code after EO when num-lock is on */

#define INSERT_KEY_SCAN_CODE        0x70    /**< Scan code for the INSERT key */
#define HOME_KEY_SCAN_CODE          0x6C    /**< Scan code for the HOME key */
#define END_KEY_SCAN_CODE           0x69    /**< Scan code for the END key */
#define PGUP_KEY_SCAN_CODE          0x7D    /**< Scan code for the PAGE UP key */
#define PGDN_KEY_SCAN_CODE          0x7A    /**< Scan code for the PAGE DOWN key */
#define PRINTSCRN_SCAN_CODE         0x7C    /**< Scan code for the PRTSC key */
#define PAUSE_SCAN_CODE_1ST         0x14    /**< Extended scan codes for PAUSE (0xE1,0x14,0x77) */
#define PAUSE_SCAN_CODE_2ND         0x77    /**< Extended scan codes for PAUSE (0xE1,0x14,0x77) */

#define ARROW_UP_KEY_SCAN_CODE      0x75    /**< Scan code for the UP ARROW key */
#define ARROW_DOWN_KEY_SCAN_CODE    0x72    /**< Scan code for the DOWN ARROW key */
#define ARROW_RIGHT_KEY_SCAN_CODE   0x6B    /**< Scan code for the RIGHT ARROW key */
#define ARROW_LEFT_KEY_SCAN_CODE    0x74    /**< Scan code for the LEFT ARROW key */



#define MAX_CMD_BYTES                  2    /**< Maximum command bytes buffered */


typedef enum
{
 PS2_RX_MODE = 0,
 PS2_TX_MODE
} PS2_TxRxModeType;

typedef enum
{
    KEY_DETECT = 0,
    KEY_READ_ID,        
    KEY_SET_SCANCODE_SET,
    KEY_WAIT_FOR_RX,
    KEY_SEND_COMMAND,             
    KEY_SEND_REQUEST_TO_SEND,
    KEY_WAIT_SEND_COMPLETE,        
    KEY_WAIT_NEXT,
    END_SEQUENCE = 255        /**< Used to end the startup sequence */  
            
} KeyStateType;

extern void __attribute__ ((weak)) KEY_CLOCK_CallBack(void);     

//const uint8_t StartUpCmdSequence[4] = {KEY_DETECT, KEY_READ_ID, KEY_SET_SCANCODE_SET, END_SEQUENCE};  /**< Sequence of KeyStateType stepped through at startup */
const uint8_t StartUpCmdSequence[3] = {KEY_DETECT, KEY_SET_SCANCODE_SET, END_SEQUENCE};  /**< Sequence of KeyStateType stepped through at startup */

const uint8_t KeyboardEchoCmd[1]    = {KBD_ECHO_CMD};    /**< Send echo command to keyboard. Keyboard should respond with the same (0xEE) */
const uint8_t ReadKeyboardID_Cmd[1] = {KBD_READ_ID_CMD}; /**< Send read ID command to keyboard,  */ 
const uint8_t SetCommandSet[2]      = {KBD_SET_SCAN_CODE_SET_CMD, KBD_SCAN_CODE_2};

typedef struct
{
    bool      clock_edge;        /**< The current clock edge level pending  */
    uint8_t   bitCount;          /**< Number of bits received in tyhe current frame */
    uint16_t  bits_in;           /**< Bit field of data bits read in current frame */ 
    bool      frameWaiting;      /**< Frame waiting signal */
    bool      frameTransmitDone; /**< Frame transmit done flag */
    bool      valid;             /**< Valid data received flag */ 
    uint8_t   s_code;            /**< Scan code read in current frame */
    
} KeyScanType;
    
    
typedef struct
{
  uint16_t  bits_in[MAX_FRAMES_BUFFERED];  
  uint8_t   scan_code[MAX_FRAMES_BUFFERED];
  uint8_t   key[MAX_FRAMES_BUFFERED];
  bool      valid[MAX_FRAMES_BUFFERED];
  
} KeyBufferType;

volatile KeyScanType KeyScan;
KeyBufferType        KeyBuffer;
uint8_t              frameCount;    /**< Number of frames received */
uint8_t              frameHead;
uint8_t              frameTail;


uint16_t             TxRxBitMask         = FRAME_RX_MASK;

PS2_TxRxModeType     PS2_TxRxMode        = PS2_RX_MODE; /**< PS2 transmit or receive mode */ 
uint16_t             sendData[MAX_CMD_BYTES];
uint8_t              lastByte;
uint16_t             sendDataCount          = 0;
uint16_t             receiveNumDataExpected = 0;
uint16_t             sendDataIndex;
bool                 ackPending          = false;  
bool                 ackRecvd            = false;

static bool          extendedKey         = false;

static uint8_t       KbdLEDs             = 0;    /** State of keyboard LEDs */
static uint8_t       ascii_code          = 0;
static uint8_t       scan_code           = 0;
static bool          last_byte           = false;
static bool          control_flag        = false;
static bool          shift_flag          = false;

static bool          caps_lock           = false;
static bool          scroll_lock         = false;
static bool          num_lock            = false;

static uint8_t       caps_lock_updatable    = true;
static uint8_t       scroll_lock_updatable  = true;
static uint8_t       num_lock_updatable     = true;

static KeyStateType  KeyBoardState          = KEY_DETECT;
static uint8_t       CmdSequenceStep        = 0;           /**< First KeyStateType */ 

static uint32_t      KeyboardCmdDelay_uS;
static bool          SentKeyboardInit = false;

static bool          PS2_keyboardPresentState = false;
static bool          EchoResponseExpected = false;
static bool          EchoResponseReceived = false;
static uint32_t      KeyboardAliveTimer;

 typedef struct
 {
    const uint8_t code;
    const uint8_t ascii_unmodified;
    const uint8_t ascii_modified;

 } ScanLookUpType;

 const ScanLookUpType ScanLookUp[] =
 {

    { 0x66, BS, BS },
 	{ 0x29, ' ', ' ' },
 	{ 0x0D, TAB, TAB },
  	{ 0x5A, CR, CR },
    { 0x76, ESC, ESC },

   	{ 0x4C, ';', ':' },	  // UK keyboard
   	{ 0x52, '\'', '@' },
   	{ 0x41, ',', '<' },
    { 0x49, '.', '>' },
	{ 0x54, '[', '{' },
   	{ 0x5B, ']', '}' },
    { 0x61, '\\', '|' },
    
    { 0x4A, '/', '?' },
    
    { 0x7C, '*', '*' },
    { 0x7B, '-', '-' },
    { 0x79, '+', '+' },
    
    { 0x05, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F1
	{ 0x06, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F2
 	{ 0x04, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F3
	{ 0x0C, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F4
	{ 0x03, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F5
	{ 0x0B, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F6
	{ 0x83, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F7
	{ 0x0A, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F8
	{ 0x01, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F19
	{ 0x09, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F10
	{ 0x78, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F11
	{ 0x07, NOT_IMPLEMENTED, NOT_IMPLEMENTED },   // F12

};

 const ScanLookUpType AlphaLookUp[] =
 {
	{ 0x1C, 'A', 'a' },
 	{ 0x32, 'B', 'b' },
 	{ 0x21, 'C', 'c' },
 	{ 0x23, 'D', 'd' },
 	{ 0x24, 'E', 'e' },
 	{ 0x2B, 'F', 'f' },
 	{ 0x34, 'G', 'g' },
 	{ 0x33, 'H', 'h' },
 	{ 0x43, 'I', 'i' },
 	{ 0x3B, 'J', 'j' },
 	{ 0x42, 'K', 'k' },
 	{ 0x4B, 'L', 'l' },
 	{ 0x3A, 'M', 'm' },
 	{ 0x31, 'N', 'n' },
 	{ 0x44, 'O', 'o' },
 	{ 0x4D, 'P', 'p' },
 	{ 0x15, 'Q', 'q' },
 	{ 0x2D, 'R', 'r' },
 	{ 0x1B, 'S', 's' },
 	{ 0x2C, 'T', 't' },
 	{ 0x3C, 'U', 'u' },
 	{ 0x2A, 'V', 'v' },
 	{ 0x1D, 'W', 'w' },
 	{ 0x22, 'X', 'x' },
 	{ 0x35, 'Y', 'y' },
 	{ 0x1A, 'Z', 'z' },
    
 };
 
 const ScanLookUpType LookUpNumbers[] =
 {
  	{ 0x16, '1', '!' },
 	{ 0x1E, '2', '"' },
 	{ 0x26, '3', '' }, // 0x23 sent (7 bit) '#')
 	{ 0x25, '4', '$' },
 	{ 0x2E, '5', '%' },
 	{ 0x36, '6', '^' },
 	{ 0x3D, '7', '&' },
 	{ 0x3E, '8', '*' },
 	{ 0x46, '9', '(' },
 	{ 0x45, '0', ')' },
 	{ 0x0E, '`', '' },
 	{ 0x4E, '-', '_' },
 	{ 0x55, '=', '+' },
 	{ 0x5D, '#', '~' },
    
 };
 
typedef struct
{
    const uint8_t code;
    const uint8_t ascii;

} LookUpNumPadType;


const LookUpNumPadType LookUpNumPad[] =
{

    { 0x4A, '/' },
    
    { 0x70, '0' },
    { 0x69, '1' },
    { 0x72, '2' },
    { 0x7A, '3' },
    { 0x6B, '4' },
    { 0x73, '5' },
    { 0x74, '6' },
    { 0x6C, '7' },
    { 0x75, '8' },
    { 0x7D, '9' },

};

/*
typedef struct
 {
    uint8_t code;
    char ascii;

 } ScanLookExtendedPadType;


const ScanLookExtendedPadType ScanLookUpExtendedKeys[] =
{
	{ 0x1F, L-GUI },
	{ 0x14, R-CTRL },
    { 0x27, R-GUI },
    { 0x11, R-ALT },
    { 0x70, INSERT },
    { 0x7D, PGUP },
    { 0x71, DELETE },
    { 0x69, END},
    { 0x7A, PGDN },
    { 0x75, U-ARROW },
    { 0x6B, L-ARROW },
    { 0x72, D-ARROW},
    { 0x74, R-ARROW },

}
*/


static void     InitialisePS2Keyboard(void);
static void     DebugFrame(void);
static void     SendCommandToKeyboard(const uint8_t *cmd, uint8_t length);
static void     ReSendLastByte(void);
static void     DecodeKeyboard(uint8_t code);
static bool     DecodeAlphaKeys(uint8_t scanCode, char *asciiCode);
static bool     DecodeSingleScanCodeKeys(uint8_t scanCode, char *asciiCode);
static bool     DecodeNumpadKeys(uint8_t scanCode, char *asciiCode);
static bool     DecodeControlKeys(uint8_t code);
static bool     ComputeOddParity(uint8_t uc);
static void     CheckOddParity(void);
static void     UpdateKbdLeds(void);
static void     SendEcho(void);
static uint16_t BuildFrame(uint8_t uc);
static void     RestartComms(PS2_TxRxModeType  Mode);
static void     CapsLock(void);
static void     ScrollLock(void);
static void     NumLock(void);
static bool     DecodeExtendedKeys(uint8_t scanCode, char *asciiCode);

/*
 * @ brief Enable PS2 interface
 */
void  EnablePS2Keyboard(void)
{
    // Initialise and set strobe inactive
    CodeOutInitialise();

    // Enable Multi Vector Configuration
    //INTCONbits.MVEC = 1;

    //    CNBI: PORT B Change Notification
    //    Priority: 1
    //    IPC2bits.CNBIP = 1;
    //    Sub Priority: 0
    //    IPC2bits.CNBIS = 0;

     /****************************************************************************
     * Clock Interrupt On Change: negative
     ***************************************************************************/
    //CNEN1bits.CN11IE = 1;    //Pin : RB10
    
    //CNPU2bits.CN16PUE = 1;
    //CNPD2bits.CN16PDE = 0;
    IPC4bits.CNIP = 7;       // Highest priority
    CNEN2bits.CN16IE = 1;    //Pin : RB10
    
    /****************************************************************************
     * Interrupt On Change: flag
     ***************************************************************************/
    //CNFBbits.CNFB11 = 0;    //Pin : RB10
    IFS1bits.CNIF = 0;        // Clear CNI interrupt flag
    /****************************************************************************
     * Interrupt On Change: config
     ***************************************************************************/
    //CNCONBbits.CNSTYLE = 1;    //Config for PORTB
    //CNCONBbits.ON = 1;         //Config for PORTB
        
    /****************************************************************************
     * Interrupt On Change: Interrupt Enable
     ***************************************************************************/
    //IFS0CLR= 1 << _IFS0_CNBIF_POSITION; //Clear CNBI interrupt flag
    //IEC0bits.CNBIE = 1; //Enable CNBI interrupt
    
    IEC1bits.CNIE = 1; //Enable CNI interrupt
    
    InitialisePS2Keyboard();

    // Setup PS2 port pins (on USB socket)
    PS2_PORT_SetInput();
    
    // Set keyboard clock pin change handler
    KEY_CLOCK_SetInterruptHandler(KeyClockISRHandler);

    PS2_TxRxMode = PS2_RX_MODE;
    // Restart capture of keyboard data
    RestartComms(PS2_RX_MODE);   
    
    // Send the initialisation command after a delay
    SentKeyboardInit = false;
    
        /* Set up sequence of commands at initialisation time */
    CmdSequenceStep = 0;           /**< First KeyStateType */ 
    KeyBoardState = StartUpCmdSequence[CmdSequenceStep];
    KeyboardCmdDelay_uS = GetTimer_uS();

}
/*
 * @ brief Disable PS2 interface
 */
void  DisablePS2Keyboard(void)
{
    /* Default IOC Interrupt Handler*/
    KEY_CLOCK_SetInterruptHandler(&KEY_CLOCK_CallBack);
    
    // Disable both edge interrupts
    //CNEN1Bbits.CNIE1B11 = 0;
    //CNEN0Bbits.CNIE0B11 = 0;            
    // Disable interrupt on change  
    CNEN2bits.CN16IE = 0;    //Pin : RB10
    IEC1bits.CNIE = 0;       // Disable CNI interrupt
}
/*
 @brief Initialise keyboard
 *
 */
static void InitialisePS2Keyboard(void)
{

    KeyScan.bitCount = 0;
    KeyScan.clock_edge = LOW_EDGE;
    KeyScan.frameWaiting = false;
    
    frameCount = 0;
    frameHead = 0;
    frameTail = 0;

    last_byte              = false;
    control_flag           = false;
    shift_flag             = false;

    caps_lock              = false;
    scroll_lock            = false;
    num_lock               = false;

    caps_lock_updatable    = true;
    scroll_lock_updatable  = true;
    num_lock_updatable     = true;

    sendDataCount          = 0;
    ackPending             = false;  
    ackRecvd               = false;
    
    PS2_keyboardPresentState = false;
    EchoResponseExpected     = false;
    EchoResponseReceived     = false;
    
    KeyboardAliveTimer = GetTimer_uS();
    
    TESTPOINT_1_SetDigitalOutput();
}
/*
 * @brief Set up keyboard, during start up sequence
 */
static void SendCommandToKeyboard(const uint8_t *cmd, uint8_t length)
{
uint8_t u;

    if (length < sizeof(sendData))
    {
        // Create command sequence
        for (u = 0; u < length; u++)
        {
            sendData[u] = BuildFrame(*cmd++);
            lastByte = sendData[u];
        }
        sendDataCount = length;
        sendDataIndex = 0;
        KeyBoardState = KEY_SEND_COMMAND;
    }
}
/*
 * @brief Resend the last byte.
 * @details Used when device sends a KBD_RESEND
 */
static void ReSendLastByte(void)
{
    sendData[0] = lastByte;
    sendDataCount = 1;
    sendDataIndex = 0;
    KeyBoardState = KEY_SEND_COMMAND;
}
/*
 * @brief Debug the frame
 */
static void DebugFrame(void)
{
#if DEBUG_KEY_DATA_FRAME == true
uint16_t temp;
bool     bit_n;
uint16_t bitMask = 0x0800;
#endif

#if DEBUG_KEY_DATA_FRAME == true

    temp = KeyBuffer.bits_in[frameTail];
    printf("Rx: ");

    for (int i = 0; i < BITS_EXPECTED; i++)
    {
        bitMask >>= 1;
        bit_n = (bool) ((temp & bitMask) ? 1 : 0);
        printf("%d", bit_n);
    }
#endif
    
#if DEBUG_KEY_CODE == true   
    printf("SC: 0x%02X\r\n", KeyBuffer.scan_code[frameTail]);
#endif
    
    frameTail = (frameTail + 1) % MAX_FRAMES_BUFFERED;
    frameCount--;
}
/*
 * @brief Decode PS/2 type keyboard
 * @details State machine handling data to and from PS2 keyboard.
 * Handling of commands set to keyboard needs some re-structuring to make it clearer
 * what is happening!
 *
 */
void PS2KeyboardTask(void)
{
uint16_t        bits_in;
static uint32_t waitTimer;
//static uint32_t KeyRepeatTimer; // Work in progress
static uint32_t AckTimeoutTimer;
//static bool     EchoCheckSent = false;

    switch(KeyBoardState)
    {
        case KEY_DETECT:
            
            receiveNumDataExpected = 0;
            SendCommandToKeyboard(KeyboardEchoCmd, sizeof(KeyboardEchoCmd));
            break;
        
        case KEY_READ_ID:
            
            receiveNumDataExpected = 2;
            SendCommandToKeyboard(ReadKeyboardID_Cmd, sizeof(ReadKeyboardID_Cmd));
            break;
            
        case KEY_SET_SCANCODE_SET:
            
            //  Send the keyboard initialisation command after a delay
            // Set up keyboard
            SendCommandToKeyboard(SetCommandSet, sizeof(SetCommandSet));
            SentKeyboardInit = true;
            break;
    
        case KEY_WAIT_FOR_RX:

                        // If key waiting
                        if (KeyScan.frameWaiting == true)
                        {
                            bits_in = KeyScan.bits_in;
                            CheckOddParity();    // Check parity
                            KeyScan.bits_in = 0; // Reset bits in
                            KeyScan.frameWaiting = false;
                               
                            if (KeyScan.valid)
                            {
                                // Save scan code
                                KeyScan.s_code = (uint8_t) (bits_in >> 1);

                                // Resend received ?
                                if ( KeyScan.s_code == KBD_RESEND )
                                {
                                    KEY_CLOCK_SetDigitalOutput();
                                    KEY_CLOCK_SetLow();
                                    ReSendLastByte();
                                    break;
                                }
                                
                                // Decode scan code
                                DecodeKeyboard(KeyScan.s_code);
                                
                                if (frameCount < MAX_FRAMES_BUFFERED)
                                {
                                    KeyBuffer.bits_in[frameHead] = bits_in;
                                    KeyBuffer.scan_code[frameHead] = KeyScan.s_code;
                                    KeyBuffer.valid[frameHead] = KeyScan.valid;
                                    KeyBuffer.key[frameHead] = ascii_code;
                                    frameCount++;    
                                    frameHead = (frameHead + 1) % MAX_FRAMES_BUFFERED; 
                                }
                            }
                            
                            KEY_CLOCK_SetDigitalOutput();
                            KEY_CLOCK_SetLow();

                            
                            // If acknowledge is pending, then clear acknowledge and pending flags
                            // ...or the "Echo" response was received
                            if (((ackPending == true) && (ackRecvd == true)) || (scan_code == KBD_ECHO_CMD))
                            {
                                ackRecvd = false;
                                ackPending = false;
                                PS2_keyboardPresentState = true;
                                
                                // End transmit and cancel acknowledge pending if all done
                                if (sendDataIndex == sendDataCount)
                                {
                                    sendDataCount = 0;
                                    //printf("Command %d complete\r\n", CmdSequenceStep);
                                    
                                    if (CmdSequenceStep != END_SEQUENCE)
                                    {
                                        // Next command in sequence...
                                        if (++CmdSequenceStep != END_SEQUENCE)
                                        {
                                            // Do the next step in the sequence
                                            KeyBoardState = StartUpCmdSequence[CmdSequenceStep];
                                            
                                            if (KeyBoardState == END_SEQUENCE)
                                            {
                                               CmdSequenceStep = END_SEQUENCE;
                                               KeyBoardState = KEY_WAIT_FOR_RX; 
                                            }
                                        }
                                    }
                                    // If its the last in the sequence KeyBoardState stays = KEY_WAIT_FOR_RX
                                    
                                    //printf("KeyBoardState %d, CmdSequenceStep %d\r\n", KeyBoardState, CmdSequenceStep);
                                }
                            }
                            
                            // If command byte to keyboard pending, send it
                            if ((sendDataCount != 0) && (ackPending == false))
                            {
                                KeyBoardState = KEY_SEND_COMMAND;
                            }
                            else
                            {
                                // If state not just changed by startup sequence
                               if (KeyBoardState == KEY_WAIT_FOR_RX)
                               {
                                    // Restart keyboard data comms
                                    RestartComms(PS2_RX_MODE); 
                               }
                            }
                        }
                        else
                        {
                            // No frame waiting   
                            
                            // Check for send timeout and no ack
                            if (((ackPending == true) && (ackRecvd == false)))
                            {
                                // If timed out waiting for acknowledge
                                if (labs(GetTimer_uS() - AckTimeoutTimer) >= (uint32_t) ACK_WAIT_TIMEOUT)
                                {
                                   // Cancel send  
                                   ackPending = false;
                                   sendDataCount = 0;
                                   //printf("\r\nTimeout waiting for ACK\r\n"); 
                                   if (KeyScan.frameTransmitDone == false)
                                   {
                                       // Abort
                                       InitialisePS2Keyboard();
                                   }
                                }
                            }
                            
                            /*
                            // Periodically send echo to keyboard and expect response from last time
                            if ((GetTimer_uS() - KeyboardAliveTimer) > KEYBOARD_ALIVE_INTERVAL)
                            {
                                KeyboardAliveTimer = GetTimer_uS();
                                
                                if (!EchoCheckSent)
                                {
                                    //puts("echo sent\r\n");
                                    SendEcho();
                                }
                                
                                if ((EchoCheckSent) && (EchoResponseReceived))
                                {
                                    puts("Keyboard active\r\n");
                                }
                                else
                                {
                                    //puts("Keyboard not active\r\n");
                                }
                                
                                EchoCheckSent = !EchoCheckSent;
                            }
                            */
                        }
                        break;
        
        case KEY_SEND_COMMAND:
                        // Initiate a host send. Request to send...
                        waitTimer = GetTimer_uS();
                        // Set clock to output and hold for >= 100uS
                        KEY_CLOCK_SetLow();
                        KEY_CLOCK_SetDigitalOutput();
                        KeyBoardState = KEY_SEND_REQUEST_TO_SEND;
                        break;
                        
        case KEY_SEND_REQUEST_TO_SEND:
            
                        // Wait while CLOCK low 
                        if (labs(GetTimer_uS() - waitTimer) >= (uint32_t) REQUEST_TO_SEND_DELAY)
                        {
                            AckTimeoutTimer = GetTimer_uS();
                            // Send next byte PS2 Transmit mode
                            RestartComms(PS2_TX_MODE); 
                            KeyBoardState = KEY_WAIT_SEND_COMPLETE;
                        }
                        break;
                        
        case KEY_WAIT_SEND_COMPLETE:
            
                        // Wait for send complete (waiting now for acknowledge)
                        if (KeyScan.frameTransmitDone)        
                        {
                            sendDataIndex++;
                            waitTimer = GetTimer_uS();
                            // Restart keyboard data comms in receive mode
                            RestartComms(PS2_RX_MODE);  
                            KeyBoardState = KEY_WAIT_FOR_RX;
                        }
                        break;

        default:
            // Should never get here
            KeyBoardState = KEY_WAIT_FOR_RX;
            break;
    }
    
    // If Debug needed
    if (frameCount != 0)
    {
        DebugFrame();
    }

}
/*
 * @brief Return state of PS2 keyboard
 */
bool PS2KeyboardConnected(void)
{
    return PS2_keyboardPresentState;
}
/*
 * @brief Key clock edge interrupt
 * @details In receive frame mode (PS2_RX_MODE) process all 11 bits
 *          In transmit mode miss out the start bit which was sent during
 *          request to send: 
            PS2_TX_MODE bits: data 0-7, parity, stop, acknowledge
 */
void KeyClockISRHandler(void)
{
uint8_t wait15uS;  

#define DELAY_FOR_BIT_SAMPLE    8 /**< Delay value for bit center with clock = 24Mhz PIC32MM */

    // Clear Interrupt On Change: flag
    IFS1bits.CNIF = 0; // Clear CNI interrupt flag

    // Delay approx 5uS for sample
    // Value chosen empirically
    wait15uS = 0;
    
    while (++wait15uS < DELAY_FOR_BIT_SAMPLE)
    {
    }

    //****************************************************    
    // Waiting for clock high edge ?
    //****************************************************    
    if (KeyScan.clock_edge == HIGH_EDGE) 
    {
        // Correct edge?
        if (KEY_CLOCK_GetValue() == 1)
        {
            // Interrupt on change negative
            CNEN2bits.CN16IE = 1;             //Pin : RB10
            KeyScan.clock_edge = LOW_EDGE;    // Wait for low edge
        }
    }
    else 
    {
        //****************************************************
        // If waiting for clock low edge
        //****************************************************
        
        // Correct edge?
        if (KEY_CLOCK_GetValue() == 0)
        {
            //********************
            // If in receive mode
            //********************
            if (PS2_TxRxMode == PS2_RX_MODE)
            {

                TESTPOINT_1_SetHigh();

                if (KEY_DATA_GetValue() == 1)
                {
                    KeyScan.bits_in |= TxRxBitMask;
                }

                TxRxBitMask <<= 1;
                KeyScan.bitCount++;

                // RX mode
                if (KeyScan.bitCount == RX_BITS_EXPECTED)
                {
                    KeyScan.bitCount = 0;
                    TxRxBitMask = FRAME_RX_MASK; 
                    KeyScan.frameWaiting = true;
                    // Disable both edge interrupts
                    CNEN2bits.CN16IE = 0;    //Pin : RB10
                   }
                else
                {
                    // Interrupt on change positive
                    CNEN2bits.CN16IE = 1; 
                }

                TESTPOINT_1_SetLow();
            }
            //********************
            // PS2 transmit mode
            //********************
            else if (PS2_TxRxMode == PS2_TX_MODE)
            {
                // Send only 8 data bits + parity and stop
                if (KeyScan.bitCount < (TX_BITS_EXPECTED -1))
                {
                    if (sendData[sendDataIndex] & TxRxBitMask)
                    {
                        KEY_DATA_SetHigh();
                    }
                    else
                    {
                        KEY_DATA_SetLow();
                    }
                }

                TxRxBitMask <<= 1;
                KeyScan.bitCount++;

                if (KeyScan.bitCount == (TX_BITS_EXPECTED -1))
                {
                    // After the stop bit
                    // Release the data line ready for the ACK from the device
                    KEY_DATA_SetDigitalInput();   
                    // Interrupt on change positive
                    CNEN2bits.CN16IE = 1;  //Pin : RB10
                }
                else if (KeyScan.bitCount == TX_BITS_EXPECTED)
                {
                    // ACK bit
                    TxRxBitMask = FRAME_TX_MASK; 
                    KeyScan.frameTransmitDone = true;
                    // Disable both edge interrupts
                    CNEN2bits.CN16IE = 0;    //Pin : RB10
                }
                else
                {
                    // Interrupt on change positive
                    CNEN2bits.CN16IE = 1;      //Pin : RB10
                }
            }
            KeyScan.clock_edge = HIGH_EDGE;    // Now wait for high edge
        }
    }

}
/*
 * @brief Update LEDs on keyboard
 * @param uint8_t *bytes, Data to send
 * @param uint8_t len Length of bytes to send
 * 
 */
static void UpdateKbdLeds(void)
{
    sendData[0] = BuildFrame(KBD_SET_RESET_LEDS_CMD);
    sendData[1] = BuildFrame(KbdLEDs);
    sendDataCount = 2;
    sendDataIndex = 0;
}
/*
 * @brief Send echo command to keyboard
 */
static void SendEcho(void)
{
    sendData[0] = BuildFrame(KBD_ECHO_CMD);
    sendDataCount = 1;
    sendDataIndex = 0;
    
    // Expect response
    EchoResponseExpected = true;
    EchoResponseReceived = false;
    KeyBoardState = KEY_SEND_COMMAND;
}
/*
 * @ brief Build data frame start, parity and stop
 * @param uint8_t uc, value 
 * @return Data to transmit 
 * 
 * Returned value
 * Bits 15-11: 0
 *  bit 10: Stop bit (1)
 *  bit  9: Parity computed (ODD)
 *  bit  8: Data bit 7
 *  bit  7: Data bit 6
 *  bit  6: Data bit 5
 *  bit  5: Data bit 4
 *  bit  4: Data bit 3
 *  bit  3: Data bit 2
 *  bit  2: Data bit 1
 *  bit  1: Data bit 0
 *  bit  0: Start bit (0)
 * 
 */
static uint16_t BuildFrame(uint8_t uc)
{
uint16_t frame;
bool parity;

    frame = 0;  
    frame |= (uint16_t) (uc << 1);
    parity = ComputeOddParity(uc);
    
    if (parity == false)
    {
        frame |= 0x0200; // Set parity bit
    }
    frame |= 0x0400;     // Set stop bit
    
    return frame;
}
/*
 * @brief Compute ODD parity
 * @param uint8_t uc, value 
 */
static bool ComputeOddParity(uint8_t uc)
{
bool parity;

    parity = 0; 

    while (uc) 
    { 
        parity = !parity; 
        uc = uc & (uc - 1); 
    }      
   
    return parity;
}
/*
 * @brief Check ODD parity
 * @return true if valid
 */
static void CheckOddParity(void) 
{ 
uint8_t  uc;
uint16_t temp;
bool     rxParityBit;
bool     parity;

    temp = KeyScan.bits_in;
    temp >>= 9;
    rxParityBit = (bool) (temp & 0x0001);
    uc = (uint8_t) (KeyScan.bits_in >> 1);
    
    parity = ComputeOddParity(uc); 
    
    // Odd parity check
    KeyScan.valid = ((parity != rxParityBit) ? true : false);
} 
/*
 * @brief Restart communications keyboard data
 * @param PS2_TxRxModeType  Mode, PS2_RX_MODE = receive
 *                                PS2_TX_MODE = transmit
 */
static void RestartComms(PS2_TxRxModeType  Mode)
{
    // Set transmit or receive mode
    PS2_TxRxMode = Mode;
    KeyScan.bitCount = 0;

    KeyScan.clock_edge = LOW_EDGE;       // Now wait for low edge
    
    if (PS2_TxRxMode == PS2_RX_MODE)
    {
        TxRxBitMask = FRAME_RX_MASK;     // Set initial data mask
        KEY_DATA_SetDigitalInput();      // Set data to input
        KEY_CLOCK_SetDigitalInput();     // Clock to input
    }
    else if (PS2_TxRxMode == PS2_TX_MODE)
    {
        ackRecvd = false;                // Reset acknowledged flag 
        ackPending = true;               // An acknowledge is pending  
        KeyScan.frameTransmitDone = false;
        TxRxBitMask = FRAME_TX_MASK;     // Set initial data mask
        KEY_DATA_SetDigitalOutput();     // Set data to output
        KEY_DATA_SetLow();               // Bring data line low (START BIT)
        KEY_CLOCK_SetDigitalInput();     // Release the clock
    }
    
    // Interrupt on change negative
   
    //CNPU2bits.CN16PUE = 1;
    //CNPD2bits.CN16PDE = 0;
    //IFS1bits.CNIF = 0;                   // Clear CNI interrupt flag
    CNEN2bits.CN16IE = 1;                // Enable ICN interrupt Pin : RB10

}
/*
 * @brief Decode key data
 * @param uint8_t code, The scan-code
 */
static void DecodeKeyboard(uint8_t code)
{
static uint8_t  lastExtendedCode = 0;  /**< Stores the last extended key scan code */   
static bool     waitExtCodeBreak_E0 = false;
static bool     waitExtCodeBreak_E1 = false;
bool found =    false;

    // Copy of current scan code
    scan_code = code;

    // If acknowledge received from host command bytes
    if (scan_code == KBD_PS2_ACK) 
    {
         // Register response to periodic echo command send to keyboard
         if (EchoResponseExpected)
         {
            EchoResponseReceived = true;
         }
         
        ackRecvd = true;
        return;
    }
    
    // Extended break code
    if ((waitExtCodeBreak_E0) && (scan_code == EX_SCAN_CODE_1))
    {
        waitExtCodeBreak_E0 = false;
        return;
    }
    
    // Break code for PAUSE key
    if ((waitExtCodeBreak_E1) && (scan_code == EX_SCAN_CODE_2))
    {
        waitExtCodeBreak_E1 = false;
        return;
    }
    
    // check for key-up code
    if (scan_code == KEY_BREAK_CMD)
    {
        last_byte = true; // set last byte flag
        return;
    }
    
    // PAGE-UP, PAGE-DOWN
    if (scan_code == EX_SCAN_CODE_1)
    {
        waitExtCodeBreak_E0 = true;
        extendedKey = true;
    }

    // Scan code prefix 0xE1 for PAUSE key
    if (scan_code == EX_SCAN_CODE_2)
    {
        waitExtCodeBreak_E1 = true;
        extendedKey = true;
    }
    
    (void) DecodeControlKeys(scan_code);
    
    if (!last_byte)
    {
        found = DecodeAlphaKeys(code, &ascii_code);
    
        if (!found)
        {
            found = DecodeSingleScanCodeKeys(code, &ascii_code);
            
            if (!found)
            {
                if (code == DELETE_KEY_SCAN_CODE)
                {
                    found = true;
                    
                    if (num_lock)
                    {
                        ascii_code = '.';   
                    }
                    else
                    {
                        ascii_code = DEL;
                    }
                }
            }
        }


        if ((found == false) && (extendedKey == false))
        {
            #if NUM_KEYS_NEED_NUM_LOCK == true
            found = DecodeNumpadKeys(code, &ascii_code);
                
            if ((!num_lock) && ((ascii_code >= '0') && (ascii_code <= '9'))) 
            {
                found = false;
            }
            #else
            found = DecodeNumpadKeys(code, &ascii_code);
            #endif
        }

        if (!found)
        {
            if (extendedKey == true)
            {
                // Save code, so we can detect extended key up
                lastExtendedCode = code;
                found = DecodeExtendedKeys(code, &ascii_code);    
            }
            else
            {
                lastExtendedCode = 0;    
            }
        }
        
        // Output valid ASCII key code to host
        if (found == true)
        {
            // Control key down ?
            if (control_flag)
            {
                if ((ascii_code >= 0x40) && (ascii_code <= 0x5F))
                {
                    ascii_code = ascii_code - 0x40;
                }
                else if ((ascii_code >= 0x60) && (ascii_code <= 0x7A))
                {
                    ascii_code = ascii_code - 0x60;
                }
            }
            SendCharToHost(ascii_code);

            #if DEBUG_ASCII_CODE == true
            printf("ASCII 0x%02X\r\n", ascii_code);
            #endif
        }
    }
    
    
    if (last_byte)
    {
        last_byte = false;
              extendedKey = false;
    }
    
}
/*
 * @brief Decode alpha keys mapped to 'A'-'Z', 'a' - 'z'
 * @param uint8_t code, The scan-code
 */
static bool DecodeAlphaKeys(uint8_t scanCode, char *asciiCode)
{
bool decoded = false;
int i;

    for ( i = 0 ; i < (sizeof(AlphaLookUp) /  sizeof(ScanLookUpType)); i++ )
    {
        if (AlphaLookUp[i].code == scanCode) 
        {
            if (( (caps_lock) || (shift_flag)) && 
               (!((caps_lock) && (shift_flag))))
            {
                *asciiCode = AlphaLookUp[i].ascii_unmodified;
                decoded = true;
            }
            else
            {
                *asciiCode = AlphaLookUp[i].ascii_modified;
                decoded = true;
            }
            
        }
    }

    return decoded;
}
/*
 * @brief Decode keys with single byte scan code
 * @param uint8_t code, The scan-code
 */
static bool DecodeSingleScanCodeKeys(uint8_t scanCode, char *asciiCode)
{
bool decoded = false;
int i;

    for ( i = 0 ; i < (sizeof(ScanLookUp) /  sizeof(ScanLookUpType)); i++ )
    {
        if (ScanLookUp[i].code == scanCode) 
        {
            if (shift_flag)
            {
                if (ScanLookUp[i].ascii_modified != NOT_IMPLEMENTED)
                {
                    *asciiCode = ScanLookUp[i].ascii_modified;
                    decoded = true;
                }
            }
            else
            {
                if (ScanLookUp[i].ascii_unmodified != NOT_IMPLEMENTED)
                {
                    *asciiCode = ScanLookUp[i].ascii_unmodified;
                    decoded = true;
                }
            }
            
        }
    }

    if (!decoded)
    {
        for ( i = 0 ; i < (sizeof(LookUpNumbers) /  sizeof(ScanLookUpType)); i++ )
        {
            if (LookUpNumbers[i].code == scanCode) 
            {
                if (shift_flag)
                {
                    *asciiCode = LookUpNumbers[i].ascii_modified;  
                }
                else
                {
                    *asciiCode = LookUpNumbers[i].ascii_unmodified;  
                }
                decoded = true;
            }
        }
    }
    
    return decoded;
}
/*
 * @brief Decode num pad keys 
 * @param uint8_t code, The scan-code
 */
static bool DecodeNumpadKeys(uint8_t scanCode, char *asciiCode)
{
bool decoded = false;
int i;

    for ( i = 0 ; i < (sizeof(LookUpNumPad) /  sizeof(LookUpNumPadType)); i++ )
    {
        if (LookUpNumPad[i].code == scanCode) 
        {
            *asciiCode = LookUpNumPad[i].ascii;
            decoded = true;
        }
    }
        
    return decoded;
}
/*
 * @brief Decode control keys
 * @param uint8_t code, The scan-code
 */
static bool DecodeControlKeys(uint8_t code)
{
bool decoded = false;    

    // left & right shift keys
    if ((extendedKey == false) && ((code == LEFT_SHIFT_KEY_SCAN_CODE) || (code == RIGHT_SHIFT_KEY_SCAN_CODE)))
    {
        // If last code was key-up
        if (last_byte)
        {
            // shift key up
            // reset shift
            shift_flag = false;
        }
        else
        {
            // set shift flag
            shift_flag = true; 
        }
    }
    else if ((extendedKey == false) && (code == LEFT_CTRL_KEY_SCAN_CODE))
    {
        if (true == last_byte)
        {
            // control key up
            // reset control flag
            control_flag = false;
        }
        else
        {
            control_flag = true;
        }
    }
    else if (code == CAPS_LOCK_SCAN_CODE)
    {
        if (true == last_byte)
        {
           caps_lock_updatable = true;
        }
        else
        {
            CapsLock(); // caps lock
        }
    }
    else if ((extendedKey == false) && (code == NUM_LOCK_SCAN_CODE))
    {
        if (true == last_byte)
        {
           num_lock_updatable = true;
        }
        else
        {
            NumLock(); // Num lock
        }
    }
    else if (code == SCROLL_LOCK_SCAN_CODE)
    {
        if (true == last_byte)
        {
            scroll_lock_updatable = true;
        }
        else
        {
            ScrollLock(); // Scroll lock
        }
    }

    return decoded;
}
/*
 *@brief Handle caps lock key
 */
static void CapsLock(void)
{
    if (caps_lock_updatable == true)
    {
        if (!caps_lock)
        {
            // toggle caps lock
            caps_lock = true;
            KbdLEDs |= PS2_KBD_CAPS_LOCK_BIT;
        }
        else
        {
            caps_lock = false;
            KbdLEDs &= ~PS2_KBD_CAPS_LOCK_BIT;
        }
    
        // Send command to set/reset LED
        UpdateKbdLeds();
        caps_lock_updatable = false;
    }
}
/*
 *@brief Handle scroll lock key
 */
static void ScrollLock(void)
{
    if (scroll_lock_updatable == true)
    {
        if (!scroll_lock)
        {
            // toggle caps lock
            scroll_lock = true;
            KbdLEDs |= PS2_KBD_SCROLL_LOCK_BIT;
        }
        else
        {
            scroll_lock = false;
            KbdLEDs &= ~PS2_KBD_SCROLL_LOCK_BIT;
        }
        // Send command to set/reset LED
        UpdateKbdLeds();
        scroll_lock_updatable = false;
    }
}
/*
 *@brief Handle num lock key
 */
static void NumLock(void)
{
    if (num_lock_updatable == true)
    {
        if (!num_lock)
        {
            // toggle caps lock
            num_lock = true;
            KbdLEDs |= PS2_KBD_NUM_LOCK_BIT;
        }
        else
        {
            num_lock = false;
            KbdLEDs &= ~PS2_KBD_NUM_LOCK_BIT;
        }
        // Send command to set/reset LED
        UpdateKbdLeds();
        num_lock_updatable = false;
    }
}
/*
 * @brief Decode extended key codes
 */
static bool DecodeExtendedKeys(uint8_t scanCode, char *asciiCode)
{
bool decoded = false;

    //**************************************************
    // Extended key codes
    //**************************************************
    if (scanCode == DELETE_KEY_SCAN_CODE)
    {
        *asciiCode = DEL;    
        decoded = true;
    }  
    else if (scanCode == PGDN_KEY_SCAN_CODE)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PGDN\r\n");
        #endif        
        extendedKey = true; 
    }
    else if (scanCode == PGUP_KEY_SCAN_CODE)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PGUP\r\n");
        #endif        
        extendedKey = true; 
    }
    else if (scanCode == HOME_KEY_SCAN_CODE)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("HOME\r\n");
        #endif        
    }
    else if (scanCode == END_KEY_SCAN_CODE)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("END\r\n");
        #endif        
    }
    else if (scanCode == INSERT_KEY_SCAN_CODE)        
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("INSERT\r\n");
        #endif        
    }
    else if (scanCode == PRINTSCRN_SCAN_CODE)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PRTSCN\r\n");
        #endif        
    }
    else if (scanCode == PAUSE_SCAN_CODE_2ND)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PAUSE\r\n");
        #endif        
    }
    else if (scanCode == ARROW_UP_KEY_SCAN_CODE)
    {
        *asciiCode = '8';    
        decoded = true;
    }
    else if (scanCode == ARROW_DOWN_KEY_SCAN_CODE)
    {
        *asciiCode = '2';    
        decoded = true;
    }
    else if (scanCode == ARROW_RIGHT_KEY_SCAN_CODE)
    {
        *asciiCode = '4';    
        decoded = true;
    }
    else if (scanCode == ARROW_LEFT_KEY_SCAN_CODE)
    {
        *asciiCode = '6';    
        decoded = true;
    }
    else
    {
        decoded = false;    
    }
    
    return decoded; 
}
// EOF