/* ************************************************************************** */
/** 

  @Company
   JW Technical Consultancy Ltd

  @File Name
    USB_keyboard.c

  @Summary
    USB keyboard

  @Description
     USB keyboard driver
    
     Based on Microchip application. Modified for handling extra keys
         
*/
/* ************************************************************************** */
#include "mcc_generated_files/system.h"
#include "mcc_generated_files/pin_manager.h"
#include "mcc_generated_files/interrupt_manager.h"
#include <stdbool.h>
#include <stdio.h>
#include <stdio.h>
#include <string.h>
#include "global.h"
#include "mcc_generated_files/usb/usb.h"
#include "Timer.h"
#include "kbd_international_settings.h"
#include "kbd_parallel_out.h"
#include "USB_keyboard.h"

#define DEBUG_KEY_CODE          false  /**< Output the key code */
#define DEBUG_UNHANDLED_KEYS    false  /**< Output serial debug info for unhandled keys */

/**
 * Modifier masks - used for the first byte in the HID report.
 * NOTE: The second byte in the report is reserved, 0x00
 */
#define KEY_MOD_LCTRL  0x01
#define KEY_MOD_LSHIFT 0x02
#define KEY_MOD_LALT   0x04
#define KEY_MOD_LMETA  0x08
#define KEY_MOD_RCTRL  0x10
#define KEY_MOD_RSHIFT 0x20
#define KEY_MOD_RALT   0x40
#define KEY_MOD_RMETA  0x80

// *****************************************************************************
// *****************************************************************************
// Type definitions
// *****************************************************************************
// *****************************************************************************

typedef enum _APP_STATE
{
    DEVICE_NOT_CONNECTED,
    WAITING_FOR_DEVICE,
    DEVICE_CONNECTED, /* Device Enumerated  - Report Descriptor Parsed */
    GET_INPUT_REPORT, /* perform operation on received report */
    INPUT_REPORT_PENDING,
    SEND_OUTPUT_REPORT, /* Not needed in case of mouse */
    OUTPUT_REPORT_PENDING,
    ERROR_REPORTED
} KEYBOARD_STATE;


typedef struct
{
    uint8_t address;
    KEYBOARD_STATE state;
    bool inUse;

    struct
    {
        uint16_t id;
        uint16_t size;
        uint16_t pollRate;
        uint8_t *buffer;

        struct
        {           
            struct
            {
                HID_DATA_DETAILS details;
                HID_USER_DATA_SIZE newData[6];
                HID_USER_DATA_SIZE oldData[6];
            } parsed;
        } normal;

        struct
        {
            struct{
                HID_USER_DATA_SIZE data[8];
                HID_DATA_DETAILS details;
            } parsed;
        } modifier;
    } keys;

    struct
    {
        bool updated;
        
        union
        {
            uint8_t value;
            struct
            {
                uint8_t  numLock       : 1;
                uint8_t  capsLock      : 1;
                uint8_t  scrollLock    : 1;
                uint8_t                : 5;
            } bits;
        } report;
        struct
        {
            HID_DATA_DETAILS details;
        } parsed;
    } leds;
} KEYBOARD;

typedef struct
{
    USB_HID_KEYBOARD_KEYPAD key;
    char unmodified;
    char modified;
} HID_KEY_TRANSLATION_TABLE_ENTRY;

#define MAX_ERROR_COUNTER               (10)

// *****************************************************************************
// *****************************************************************************
// Local Variables
// *****************************************************************************
// *****************************************************************************
static usbKeyboardConnected = false;

static KEYBOARD keyboard;

static const HID_KEY_TRANSLATION_TABLE_ENTRY keyTranslationTable[] = 
{
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_A,                                         'a', 'A' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_B,                                         'b', 'B' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_C,                                         'c', 'C' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_D,                                         'd', 'D' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_E,                                         'e', 'E' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_F,                                         'f', 'F' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_G,                                         'g', 'G' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_H,                                         'h', 'H' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_I,                                         'i', 'I' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_J,                                         'j', 'J' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_K,                                         'k', 'K' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_L,                                         'l', 'L' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_M,                                         'm', 'M' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_N,                                         'n', 'N' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_O,                                         'o', 'O' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_P,                                         'p', 'P' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_Q,                                         'q', 'Q' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_R,                                         'r', 'R' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_S,                                         's', 'S' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_T,                                         't', 'T' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_U,                                         'u', 'U' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_V,                                         'v', 'V' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_W,                                         'w', 'W' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_X,                                         'x', 'X' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_Y,                                         'y', 'Y' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_Z,                                         'z', 'Z' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_1_AND_EXCLAMATION_POINT,                   '1', '!' },
    //{ USB_HID_KEYBOARD_KEYPAD_KEYBOARD_2_AND_AT,                                  '2', '@' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_2_AND_AT,                                  '2', '\"' }, // Logitech and Mels'
#if UK_POUND_CODE == true
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_3_AND_HASH,                                '3', '#' }, 
#else
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_3_AND_HASH,                                '3', '#' },
#endif
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_4_AND_DOLLAR,                              '4', '$' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_5_AND_PERCENT,                             '5', '%' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_6_AND_CARROT,                              '6', '^' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_7_AND_AMPERSAND,                           '7', '&' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_8_AND_ASTERISK,                            '8', '*' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_9_AND_OPEN_PARENTHESIS,                    '9', '(' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_0_AND_CLOSE_PARENTHESIS,                   '0', ')' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_MINUS_AND_UNDERSCORE,                      '-', '_' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_EQUAL_AND_PLUS,                            '=', '+' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_OPEN_BRACKET_AND_OPEN_CURLY_BRACE,         '[', '{' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_CLOSE_BRACKET_AND_CLOSE_CURLY_BRACE,       ']', '}' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_BACK_SLASH_AND_PIPE,                       '\\', '|' },
    //{ USB_HID_KEYBOARD_KEYPAD_KEYBOARD_NON_US_HASH_AND_TILDE,                     '`', '~' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_NON_US_HASH_AND_TILDE,                     '#', '~' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_SEMICOLON_AND_COLON,                       ';', ':' },
    //{ USB_HID_KEYBOARD_KEYPAD_KEYBOARD_APOSTROPHE_AND_QUOTE,                      '\'', '\"' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_NON_US_FORWARD_SLASH_AND_PIPE,             '\\', '|' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_APOSTROPHE_AND_QUOTE,                      '\'', '@' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_GRAVE_ACCENT_AND_TILDE,                    '`', '~' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_COMMA_AND_LESS_THAN,                       ',', '<' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_PERIOD_AND_GREATER_THAN,                   '.', '>' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_FORWARD_SLASH_AND_QUESTION_MARK,           '/', '?' },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_SPACEBAR,                                  ' ', ' ' },
    
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_DELETE_FORWARD,                            DEL, DEL },
    { USB_HID_KEYBOARD_KEYPAD_KEYBOARD_TAB,                                       TAB, TAB },
    
};

// Escape sequences 
#define END  0x00
#define HOME 0x00

#define SPECIAL  0x00

static const HID_KEY_TRANSLATION_TABLE_ENTRY NumPadKeyTranslationTable[] = 
{
    
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_BACK_SLASH,                                  '/', '/' },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_ASTERISK,                                    '*', '*' },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_MINUS,                                       '-', '-' },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_PLUS,                                        '+', '+' },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_ENTER,                                        CR, CR },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_1_AND_END,                                   '1', END },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_2_AND_DOWN_ARROW,                            '2', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_3_AND_PAGE_DOWN,                             '3', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_4_AND_LEFT_ARROW,                            '4', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_5,                                           '5', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_6_AND_RIGHT_ARROW,                           '6', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_7_AND_HOME,                                  '7', HOME },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_8_AND_UP_ARROW,                              '8', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_9_AND_PAGE_UP,                               '9', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_0_AND_INSERT,                                '0', SPECIAL },
    { USB_HID_KEYBOARD_KEYPAD_KEYPAD_PERIOD_AND_DELETE,                           '.', DEL },
    
    
    //USB_HID_KEYBOARD_KEYPAD_KEYPAD_COMMA                                     = 0x85,
    //USB_HID_KEYBOARD_KEYPAD_KEYPAD_EQUAL_SIGN                                = 0x86,

    
};

const char *GREEN_BLACK = "20;40m"; /**< ANSI code for foreground green, background black */
const char *WHITE_BLACK = "37;40m"; /**< ANSI code for foreground white, background black */

#define USB_KEYBOARD_REPORT_INTERVAL_mS      10 /**< Keyboard report request interval in milli seconds */

uint8_t ascii_code = 0;
uint8_t lastKey = 0; // For auto key repeat
uint8_t keyReleased = true;
uint32_t keyTimer;
uint32_t keyRepeatTime = KEY_REPEAT_INITIAL_TIME_uS;

static uint32_t AppTimer;

extern uint32_t elapsedTime_mS(uint32_t since);

// *****************************************************************************
// *****************************************************************************
// Local Function Prototypes
// *****************************************************************************
// *****************************************************************************
static void App_ProcessInputReport(void);
static bool CursorKeys(HID_USER_DATA_SIZE key);
static bool NumPadCursorKeys(HID_USER_DATA_SIZE key);
static void SendAsciiEscapeChar(const char ch);
static void SendAsciiEscapeString(const char *str);

// *****************************************************************************
// *****************************************************************************
// Functions
// *****************************************************************************
// *****************************************************************************

/*********************************************************************
* Function: void APP_HostHIDTimerHandler(void);
*
* Overview: Switches over the state machine state to get a new report
*           periodically if the device is idle
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
static void APP_HostHIDTimerHandler(void)
{
    if (keyboard.state == DEVICE_CONNECTED)
    {
        keyboard.state = GET_INPUT_REPORT;
    }
}

/*********************************************************************
* Function: void APP_HostHIDKeyboardInitialize(void);
*
* Overview: Initializes the demo code
*
* PreCondition: None
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_HostHIDKeyboardInitialize()
{
    usbKeyboardConnected = false;
    keyboard.state = DEVICE_NOT_CONNECTED;
    keyboard.inUse = false;
    keyboard.keys.buffer = NULL;
    keyboard.address = 0;
    AppTimer = GetTimer_uS();
}

/*********************************************************************
* Function: void APP_HostHIDKeyboardTasks(void);
*
* Overview: Keeps the demo running.
*
* PreCondition: The demo should have been initialized via
*   the APP_HostHIDKeyboardInitialize()
*
* Input: None
*
* Output: None
*
********************************************************************/
void APP_HostHIDKeyboardTasks()
{
uint8_t error;
uint8_t count;
uint8_t deviceStatus;
static bool firstEnumerated = false;
    
    if (keyboard.address == 0)
    {
        keyboard.address = USBHostHIDDeviceDetect();
    }
    else
    {
        if (USBHostHIDDeviceStatus(keyboard.address) == USB_HID_DEVICE_NOT_FOUND)
        {
            printf("HID detached\r");
            usbKeyboardConnected = false;
            keyboard.state = DEVICE_NOT_CONNECTED;
            keyboard.address = 0;
            keyboard.inUse = false;

            if (keyboard.keys.buffer != NULL)
            {
                free(keyboard.keys.buffer);
                keyboard.keys.buffer = NULL;
            }
        }      
        else
        {
            if (firstEnumerated == false)
            {
                firstEnumerated = true;
                AppTimer = GetTimer_uS();
            }
            
            // Timer to request input report from keyboard
            if ( elapsedTime_mS(AppTimer) >= (USB_KEYBOARD_REPORT_INTERVAL_mS * 10) )
            {
                AppTimer = GetTimer_uS();
                APP_HostHIDTimerHandler();
            }
        }
    }

    switch(keyboard.state)
    {
        case DEVICE_NOT_CONNECTED:
            
            memset(&keyboard.keys, 0x00, sizeof(keyboard.keys));
            memset(&keyboard.leds, 0x00, sizeof(keyboard.leds));
            keyboard.state = WAITING_FOR_DEVICE;
            break;
            
        case WAITING_FOR_DEVICE:
            
            if ( (keyboard.address != 0) &&
                 ((deviceStatus = USBHostHIDDeviceStatus(keyboard.address)) == USB_HID_NORMAL_RUNNING)
               ) /* True if report descriptor is parsed with no error */
            {
                puts("USB keyboard attached\r");
                usbKeyboardConnected = true;
                keyboard.state = DEVICE_CONNECTED;
            }
            break;
            
        case DEVICE_CONNECTED:
            break;

        case GET_INPUT_REPORT:
            
            if (USBHostHIDRead(
                                keyboard.address,
                                keyboard.keys.id,
                                keyboard.keys.normal.parsed.details.interfaceNum,
                                keyboard.keys.size,
                                keyboard.keys.buffer
                             )
              )
            {
                /* Host may be busy/error -- keep trying */
            }
            else
            {
                
                keyboard.state = INPUT_REPORT_PENDING;
            }
            break;

        case INPUT_REPORT_PENDING:
            
            if (USBHostHIDReadIsComplete(keyboard.address, &error, &count))
            {
                if (error || (count == 0))
                {   
                    keyboard.state = DEVICE_CONNECTED;
                }
                else
                {
                    keyboard.state = DEVICE_CONNECTED;

                    App_ProcessInputReport();
                    
                    if (keyboard.leds.updated == true)
                    {
                        keyboard.state = SEND_OUTPUT_REPORT;
                    }
                }
            }
            break;

        case SEND_OUTPUT_REPORT: /* Will be done while implementing Keyboard */
            
            if (USBHostHIDWrite(    
                                keyboard.address,
                                keyboard.leds.parsed.details.reportID,
                                keyboard.leds.parsed.details.interfaceNum,
                                keyboard.leds.parsed.details.reportLength,
                                (uint8_t*)&keyboard.leds.report
                               )
              )
            {
                /* Host may be busy/error -- keep trying */
            }
            else
            {
                keyboard.state = OUTPUT_REPORT_PENDING;
            }
            break;

        case OUTPUT_REPORT_PENDING:
            
            if (USBHostHIDWriteIsComplete(keyboard.address, &error, &count))
            {
                keyboard.leds.updated = false;
                keyboard.state = DEVICE_CONNECTED;
            }
            break;

        case ERROR_REPORTED:
            break;

        default:
            break;

    }
}
/*
 * @brief Return true if USB keyboard connect3ed
 */
bool APP_HostHIDKeyboardConnected(void)
{
    return usbKeyboardConnected;
}
/****************************************************************************
  Function:
    BOOL USB_HID_DataCollectionHandler(void)
  Description:
    This function is invoked by HID client , purpose is to collect the
    details extracted from the report descriptor. HID client will store
    information extracted from the report descriptor in data structures.
    Application needs to create object for each report type it needs to
    extract.
    For ex: HID_DATA_DETAILS keyboard.keys.modifier.details;
    HID_DATA_DETAILS is defined in file usb_host_hid_appl_interface.h
    Each member of the structure must be initialized inside this function.
    Application interface layer provides functions :
    USBHostHID_ApiFindBit()
    USBHostHID_ApiFindValue()
    These functions can be used to fill in the details as shown in the demo
    code.

  Precondition:
    None

  Parameters:
    None

  Return Values:
    true    - If the report details are collected successfully.
    false   - If the application does not find the the supported format.

  Remarks:
    This Function name should be entered in the USB configuration tool
    in the field "Parsed Data Collection handler".
    If the application does not define this function , then HID cient
    assumes that Application is aware of report format of the attached
    device.
***************************************************************************/
bool APP_HostHIDKeyboardReportParser(void)
{
    uint8_t NumOfReportItem = 0;
    uint8_t i;
    USB_HID_ITEM_LIST* pitemListPtrs;
    USB_HID_DEVICE_RPT_INFO* pDeviceRptinfo;
    HID_REPORTITEM *reportItem;
    HID_USAGEITEM *hidUsageItem;
    uint8_t usageIndex;
    uint8_t reportIndex;
    bool foundLEDIndicator = false;
    bool foundModifierKey = false;
    bool foundNormalKey = false;

    /* The keyboard is already in use. */
    if (keyboard.inUse == true)
    {
        return false;
    }

    pDeviceRptinfo = USBHostHID_GetCurrentReportInfo(); // Get current Report Info pointer
    pitemListPtrs = USBHostHID_GetItemListPointers();   // Get pointer to list of item pointers

    /* Find Report Item Index for Modifier Keys */
    /* Once report Item is located , extract information from data structures provided by the parser */
    NumOfReportItem = pDeviceRptinfo->reportItems;
    for(i=0;i<NumOfReportItem;i++)
    {
        reportItem = &pitemListPtrs->reportItemList[i];
        if ((reportItem->reportType==hidReportInput) && (reportItem->dataModes == HIDData_Variable)&&
           (reportItem->globals.usagePage==USB_HID_USAGE_PAGE_KEYBOARD_KEYPAD))
        {
            /* We now know report item points to modifier keys */
            /* Now make sure usage Min & Max are as per application */
            usageIndex = reportItem->firstUsageItem;
            hidUsageItem = &pitemListPtrs->usageItemList[usageIndex];
            if ((hidUsageItem->usageMinimum == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_LEFT_CONTROL)
                &&(hidUsageItem->usageMaximum == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_RIGHT_GUI)) //else application cannot suuport
            {
               reportIndex = reportItem->globals.reportIndex;
               keyboard.keys.modifier.parsed.details.reportLength = (pitemListPtrs->reportList[reportIndex].inputBits + 7)/8;
               keyboard.keys.modifier.parsed.details.reportID = (uint8_t)reportItem->globals.reportID;
               keyboard.keys.modifier.parsed.details.bitOffset = (uint8_t)reportItem->startBit;
               keyboard.keys.modifier.parsed.details.bitLength = (uint8_t)reportItem->globals.reportsize;
               keyboard.keys.modifier.parsed.details.count=(uint8_t)reportItem->globals.reportCount;
               keyboard.keys.modifier.parsed.details.interfaceNum= USBHostHID_ApiGetCurrentInterfaceNum();
               foundModifierKey = true;
            }

        }
        else if ((reportItem->reportType==hidReportInput) && (reportItem->dataModes == HIDData_Array)&&
           (reportItem->globals.usagePage==USB_HID_USAGE_PAGE_KEYBOARD_KEYPAD))
        {
           /* We now know report item points to modifier keys */
           /* Now make sure usage Min & Max are as per application */
           usageIndex = reportItem->firstUsageItem;
           hidUsageItem = &pitemListPtrs->usageItemList[usageIndex];
           reportIndex = reportItem->globals.reportIndex;
           keyboard.keys.normal.parsed.details.reportLength = (pitemListPtrs->reportList[reportIndex].inputBits + 7)/8;
           keyboard.keys.normal.parsed.details.reportID = (uint8_t)reportItem->globals.reportID;
           keyboard.keys.normal.parsed.details.bitOffset = (uint8_t)reportItem->startBit;
           keyboard.keys.normal.parsed.details.bitLength = (uint8_t)reportItem->globals.reportsize;
           keyboard.keys.normal.parsed.details.count = (uint8_t)reportItem->globals.reportCount;
           keyboard.keys.normal.parsed.details.interfaceNum = USBHostHID_ApiGetCurrentInterfaceNum();
           foundNormalKey = true;
        }
        else if ((reportItem->reportType==hidReportOutput) &&
                (reportItem->globals.usagePage==USB_HID_USAGE_PAGE_LEDS))
        {
            usageIndex = reportItem->firstUsageItem;
            hidUsageItem = &pitemListPtrs->usageItemList[usageIndex];

            reportIndex = reportItem->globals.reportIndex;
            keyboard.leds.parsed.details.reportLength = (pitemListPtrs->reportList[reportIndex].outputBits + 7)/8;
            keyboard.leds.parsed.details.reportID = (uint8_t)reportItem->globals.reportID;
            keyboard.leds.parsed.details.bitOffset = (uint8_t)reportItem->startBit;
            keyboard.leds.parsed.details.bitLength = (uint8_t)reportItem->globals.reportsize;
            keyboard.leds.parsed.details.count = (uint8_t)reportItem->globals.reportCount;
            keyboard.leds.parsed.details.interfaceNum = USBHostHID_ApiGetCurrentInterfaceNum();
            foundLEDIndicator = true;
        }
    }

    if (pDeviceRptinfo->reports == 1)
    {
        keyboard.keys.id = 0;
        keyboard.keys.size = keyboard.keys.normal.parsed.details.reportLength;
        keyboard.keys.buffer = (uint8_t*)malloc(keyboard.keys.size);
        keyboard.keys.pollRate = pDeviceRptinfo->reportPollingRate;

        if ( (foundNormalKey == true) &&
            (foundModifierKey == true) &&
            (keyboard.keys.buffer != NULL)
        )
        {
            keyboard.inUse = true;
        }
    }

    return(keyboard.inUse);
}

/****************************************************************************
  Function:
    void App_ProcessInputReport(void)

  Description:
    This function processes input report received from HID device.

  Precondition:
    None

  Parameters:
    None

  Return Values:
    None

  Remarks:
    None
***************************************************************************/
static void App_ProcessInputReport(void)
{
uint8_t  i, j;
bool heldCharacter = false;
bool shift = false;
bool control_key = false;
bool found = false;
static bool numLock = false;

   /* process input report received from device */
    USBHostHID_ApiImportData(   keyboard.keys.buffer,
                                keyboard.keys.size,
                                keyboard.keys.modifier.parsed.data,
                                &keyboard.keys.modifier.parsed.details
                            );
    
    USBHostHID_ApiImportData(   keyboard.keys.buffer,
                                keyboard.keys.size,
                                keyboard.keys.normal.parsed.newData,
                                &keyboard.keys.normal.parsed.details
                            );

    if (keyboard.keys.modifier.parsed.data[USB_HID_KEYBOARD_KEYPAD_KEYBOARD_LEFT_SHIFT - USB_HID_KEYBOARD_KEYPAD_KEYBOARD_LEFT_CONTROL] == 1)
    {
        shift = true;
    }

    if (keyboard.keys.modifier.parsed.data[USB_HID_KEYBOARD_KEYPAD_KEYBOARD_RIGHT_SHIFT - USB_HID_KEYBOARD_KEYPAD_KEYBOARD_LEFT_CONTROL] == 1)
    {
        shift = true;
    }
    
            
    if ( keyboard.keys.modifier.parsed.data[0] &  KEY_MOD_LCTRL)
    {
        control_key = true;
    }
    
    if ( keyboard.keys.modifier.parsed.data[4] & KEY_MOD_RCTRL )
    {
        control_key = true;
    }

    lastKey = 0;
    
    for (i=0; i < keyboard.keys.normal.parsed.details.reportLength ;i++)
    {
        heldCharacter = false;
       
        if (keyboard.keys.normal.parsed.newData[i] != 0)
        {
            for (j=0; j < keyboard.keys.normal.parsed.details.reportLength ;j++)
            {
                if (keyboard.keys.normal.parsed.newData[i] == keyboard.keys.normal.parsed.oldData[j])
                {
                    heldCharacter = true;
                    break;
                }
            }
            
            if (heldCharacter == true)
            {
                continue;
            }
            if (keyboard.keys.normal.parsed.newData[i] == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_SCROLL_LOCK)
            {
                keyboard.leds.report.bits.scrollLock ^= 1;
                keyboard.leds.updated = true;
            }
            else if (keyboard.keys.normal.parsed.newData[i] == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_CAPS_LOCK)
            {
                keyboard.leds.report.bits.capsLock ^= 1;
                keyboard.leds.updated = true;
            }
            else if (keyboard.keys.normal.parsed.newData[i] == USB_HID_KEYBOARD_KEYPAD_KEYPAD_NUM_LOCK_AND_CLEAR)
            {
                keyboard.leds.report.bits.numLock ^= 1;
                numLock ^= 1;
                keyboard.leds.updated = true;
                ascii_code = 0;
            }
            else
            {
                HID_USER_DATA_SIZE key = keyboard.keys.normal.parsed.newData[i];
                
                //keyReleased = false;
                //keyTimer = GetTimer_uS();
                //keyRepeatTime = KEY_REPEAT_INITIAL_TIME_uS;
                
                #if DEBUG_KEY_CODE == true
                printf("   Code: 0x%02X\r\n", key);
                #endif
               
                if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_ESCAPE)
                {
                    SendCharToHost(ascii_code = ESC);
                }
                else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_RETURN_ENTER)
                {
                    SendCharToHost(ascii_code = CR);
                }
                else if ( (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_RIGHT_SHIFT) ||
                          (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_LEFT_SHIFT))
                {
                    shift = true;
                }
                else if ( key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_BACKSPACE )

                {
                    SendCharToHost(ascii_code = BS);
                }
                else if ( key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_DELETE )
                {
                    SendCharToHost(ascii_code = DEL);
                }
                else if ( (key >= USB_HID_KEYBOARD_KEYPAD_KEYBOARD_A) &&
                          (key <= USB_HID_KEYBOARD_KEYPAD_KEYBOARD_Z)
                       )
                {
                    uint8_t index;

                    for (index = 0; index < sizeof(keyTranslationTable)/sizeof(HID_KEY_TRANSLATION_TABLE_ENTRY); index++)
                    {
                        if (keyTranslationTable[index].key == key)
                        {
                            if (((keyboard.leds.report.bits.capsLock == 1) || (shift == true)) && 
                              (!((keyboard.leds.report.bits.capsLock == 1) && (shift == true))) )
                            {
                                ascii_code = keyTranslationTable[index].modified;
                            }
                            else
                            {
                                ascii_code = keyTranslationTable[index].unmodified;
                            }
                            
                            // Control key down ?
                            if (control_key)
                            {
                                if ((ascii_code >= 0x40) && (ascii_code <= 0x5F))
                                {
                                    ascii_code = ascii_code - 0x40;
                                }
                                else if ((ascii_code >= 0x61) && (ascii_code <= 0x7E))
                                {
                                    ascii_code = ascii_code - 0x60;
                                }
                            }
                
                            SendCharToHost(ascii_code);
                        }
                    }
                }
                else
                {
                    uint8_t index;

                    for (index = 0; index < sizeof(keyTranslationTable)/sizeof(HID_KEY_TRANSLATION_TABLE_ENTRY); index++)
                    {
                        if (keyTranslationTable[index].key == key)
                        {
                            found = true;
                            
                            if ( shift == true )
                            {
                                ascii_code = keyTranslationTable[index].modified;
                            }
                            else
                            {
                                ascii_code = keyTranslationTable[index].unmodified;
                            }
                            
                            // Control key down ?
                            if (control_key)
                            {
                                if ((ascii_code >= '[') && (ascii_code <= '_'))
                                {
                                    ascii_code = ascii_code - 0x40;
                                }
                            }
                            
                            SendCharToHost(ascii_code);
                        }
                    }
                    
                    if (!found)
                    {
                        for (index = 0; index < sizeof(NumPadKeyTranslationTable)/sizeof(HID_KEY_TRANSLATION_TABLE_ENTRY); index++)
                        {
                            if (NumPadKeyTranslationTable[index].key == key)
                            {
                                if ( numLock == true )
                                {
                                    SendCharToHost(ascii_code = NumPadKeyTranslationTable[index].unmodified);
                                    found = true;
                                    break;
                                }
                                else
                                {
                                    if (NumPadKeyTranslationTable[index].modified != SPECIAL)
                                    {
                                        SendCharToHost(ascii_code = NumPadKeyTranslationTable[index].modified);
                                        found = true;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    
                    if (!found)
                    {
                        found = CursorKeys(key);
                    }
                    
                    if (!found)
                    {
                        found = NumPadCursorKeys(key);
                    }
                    
                    if (!found)
                    {
                        #if DEBUG_UNHANDLED_KEYS == true
                        printf("   Un-handled Code: %02X\r\n", key);
                        #endif 
                    }
                }
                
                lastKey = ascii_code; /* For auto repeat */
                keyReleased = false;
                keyTimer = GetTimer_uS();
                keyRepeatTime = KEY_REPEAT_INITIAL_TIME_uS;
            }
        }
        else
        {
            if ( (lastKey == 0) && (heldCharacter == 0) )
            {
                keyReleased = true;
            }
            break;
        }
    }

    memcpy(keyboard.keys.normal.parsed.oldData, keyboard.keys.normal.parsed.newData, sizeof(keyboard.keys.normal.parsed.newData));
    memset(keyboard.keys.normal.parsed.newData, 0x00, sizeof(keyboard.keys.normal.parsed.newData));
}
/*
 * @brief Handle cursor keys
 */
static bool CursorKeys(HID_USER_DATA_SIZE key)
{
bool decoded = true;

    if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_ERROR_ROLL_OVER)
    {
        
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_HOME)
    {
        SendAsciiEscapeString("\x0G");
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_END)
    {
        //SendCharToHost(EOT); 
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_UP_ARROW) 
    {
        //SendAsciiEscapeChar('A');  
        SendCharToHost(ascii_code = '8');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_DOWN_ARROW) 
    {
        //SendAsciiEscapeChar('B');    
        SendCharToHost(ascii_code = '2');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_RIGHT_ARROW)
    {
        //SendAsciiEscapeChar('C');    
        SendCharToHost(ascii_code = '6');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_LEFT_ARROW)
    {
        //SendAsciiEscapeChar('D');    
        SendCharToHost(ascii_code = '4');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_INSERT)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("INSERT");
        #endif  
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_PAGE_UP)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PGUP");
        #endif  
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYBOARD_PAGE_DOWN)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PGDN");
        #endif  
    }
    else
    {
        decoded = false;
    }

    //if ( decoded )
    //{
//        lastKey = asciikey;
//        keyReleased = false;
//        keyTimer = GetTimer_uS();
//        keyRepeatTime = KEY_REPEAT_INITIAL_TIME_uS;
  //  }

    return decoded;
    
}
/*
 * @brief Handle num-pad cursor keys
 */
static bool NumPadCursorKeys(HID_USER_DATA_SIZE key)
{
bool decoded = true;

    if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_8_AND_UP_ARROW)
    {
        //SendAsciiEscapeChar('A');  
        SendCharToHost(ascii_code = '8');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_2_AND_DOWN_ARROW)
    {
        //SendAsciiEscapeChar('B');    
        SendCharToHost(ascii_code = '2');
    }   
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_5)
    {
        // Do nothing
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_6_AND_RIGHT_ARROW)
    {
        //SendAsciiEscapeChar('C');   
        SendCharToHost(ascii_code = '6');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_4_AND_LEFT_ARROW)
    {
        //SendAsciiEscapeChar('D');   
        SendCharToHost(ascii_code = '4');
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_7_AND_HOME)
    {
        SendAsciiEscapeString("\x00G");   
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_1_AND_END)
    {
        SendCharToHost(ETX);
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_9_AND_PAGE_UP)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PGUP");
        #endif  
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_3_AND_PAGE_DOWN)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("PGDN");   
        #endif  
    }
    else if (key == USB_HID_KEYBOARD_KEYPAD_KEYPAD_0_AND_INSERT)
    {
        #if DEBUG_UNHANDLED_KEYS == true
        puts("INSERT"); 
        #endif  
    }
    else
    {
        decoded = false;
    }

/*
    if ( decoded )
    {
        lastKey = asciikey;
        keyReleased = false;
        keyTimer = GetTimer_uS();
        keyRepeatTime = KEY_REPEAT_INITIAL_TIME_uS;
    }
*/
    return decoded;
}
/*
 * @brief Send ASCII escape char to parallel port
 */
static void SendAsciiEscapeChar(const char ch)
{
char seq[4];

    seq[0] = '\e';
    seq[1] = '[';
    seq[2] = ch;
    seq[3] = 0;
    
    SendStringToHost(seq, false);    
}
static void SendAsciiEscapeString(const char *str)
{
char seq[16];

    if (strlen(str) <= (sizeof(seq) + 3))
    {
        seq[0] = '\e';
        seq[1] = '[';
        strcpy(&seq[2], str);
        seq[2 + strlen(str)] = 0;

        SendStringToHost(seq, false);    
    }
}
// EOF