/*
 * kbd_parallel_out.c
 *
 * Author: JW Technical Consultancy Ltd
 * 
 * Outputs 8 bit key data to the Interak LPT1 board 
 * However LKPT1 implements only 7 bits. The 8th being used for the key data waiting strobe.
 * 
 * Outputs acquired keystroke ASCII code to keyboard parallel port.
 * Uses a 1uS timer (TMR1) to create keyboard strobe.
 * The ascii key code can be sent out via the serial interface if SERIAL_OUTPUT is set true.
 *
 */
#include "mcc_generated_files/system.h"
#include "mcc_generated_files/pin_manager.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include "global.h"
#include "Timer.h"
#include "kbd_parallel_out.h"

#define DEBUG_ASCII_CODE       false  /**< Debug ASCII code ? */

#define ONLY_PARALLEL_PORT       1
#define PARALLEL_AND_SERIAL      0

#define STROBE_HIGH              1
#define STROBE_LOW               0
#define KEYBOARD_STROBE_LEVEL    STROBE_HIGH /**< Set strobe output pulse polarity */

#define SERIAL_OUTPUT            true  /**< Enable serial output */
#define LATCH_DELAY_uS             2   /**< Latch delay in 100's of uS for parallel port */
#define STROBE_PULSE_WIDTH_uS      2   /**< Strobe pulse width in 100's of uS for parallel port */
#define PARALLEL_MAX_RATE_uS     500   /**< Maximum rate in 100's of uS for parallel port */

#define KEY_REPEAT_RATE_uS      1000   /**< Key repeat rate in 100's uS */ 
#define KEY_REPEAT_TIMEOUT_uS   4000

typedef enum
{
   CODE_OUT_IDLE = 0,
   CODE_OUT_SEND_DEBUG_CHAR,
   CODE_OUT_SEND_DEBUG_LF,
   CODE_OUT_SEND,
   CODE_LATCH_DELAY,
   CODE_STROBE_ACTIVE,
   CODE_OUT_MAX_RATE_DELAY

} CodeOutputStateType;

typedef struct
{
    char CodeOut;            /**< Data byte */
    bool OnlyToParallelPort; /**< When set ascii code out and messages go only to parallel port */
} CodeOutBufType;

#define CMD_OUT_BUFFER_SIZE  64  /**< Maximumber of buffered ascii key codes,
                                      Keyboard error messages can be reported to host */


static CodeOutputStateType  CodeOutputState = CODE_OUT_IDLE;   /**< State machine to output ascii code */
static CodeOutBufType       CodeOutBuf[CMD_OUT_BUFFER_SIZE];   /**< Buffer for dat to hardware output, with serial option */
static uint8_t              CodeOutHead  = 0;                  /**< Ascii scan code output input pointer */
static uint8_t              CodeOutTail  = 0;                  /**< Ascii scan code output Output pointer */
static uint8_t              CodeOutCount = 0;

static uint8_t  lastChar = 0;                                  /**< Copy of last ASCII character sent to host */  
static uint8_t  keyRepeatActive = false;
static uint32_t keyRepeatTimer = 0;
static uint32_t keyResendTimeout = 0;                          /**< uS timer variable for resend timeout */  

/*
 * @brief Initialise Output task
 */
void CodeOutInitialise(void)
{
    CodeOutputState = CODE_OUT_IDLE;  
    CodeOutHead  = 0; 
    CodeOutTail  = 0;  
    CodeOutCount = 0;

    STROBE_OUT_SetDigitalOutput();
    PARALLEL_KEY_DATA_SetOutput();

    #if KEYBOARD_STROBE_LEVEL == STROBE_LOW
    STROBE_OUT_SetHigh();          // Strobe - normally high - pulse low
    #elif  KEYBOARD_STROBE_LEVEL == STROBE_HIGH
    STROBE_OUT_SetLow();           // Set strobe low
    #endif
}
/*
 * @brief Task to output queued data to host
 */
void CodeOutputTask(void)
{
static uint8_t  data;
static uint32_t latchTimer;
static uint32_t ParallelPortRateTimer;

    switch (CodeOutputState)
    {
        case CODE_OUT_IDLE:

            if (CodeOutCount != 0)
            {
                // Fetch next byte to send and increment tail index
                data = CodeOutBuf[CodeOutTail].CodeOut;

                #if SERIAL_OUTPUT == true
                if (CodeOutBuf[CodeOutTail].OnlyToParallelPort == PARALLEL_AND_SERIAL)
                {
                    CodeOutputState = CODE_OUT_SEND_DEBUG_CHAR;
                }
                else
                {
                    CodeOutputState = CODE_OUT_SEND;
                }
                #else
                CodeOutputState = CODE_OUT_SEND;
                #endif

                // Next...
                CodeOutTail++;
                CodeOutTail &= (CMD_OUT_BUFFER_SIZE - 1);
                CodeOutCount--;
            }
            break;


        case CODE_OUT_SEND_DEBUG_CHAR:

            // Send data
            if (UART1_is_tx_ready())
            {
                // Don't wait for ready
                U1TXREG = data;

                if (data == CR)
                {
                    CodeOutputState = CODE_OUT_SEND_DEBUG_LF;
                }
                else
                {
                    CodeOutputState = CODE_OUT_SEND;
                }
            }
            break;

        case CODE_OUT_SEND_DEBUG_LF:

            // Send line feed
            if (UART1_is_tx_ready())
            {
                // Don't wait for ready
                U1TXREG = LF;
                CodeOutputState = CODE_OUT_SEND;
            }
            break;

            case CODE_OUT_SEND:
            // Set up data
                            
            PARALLEL_KEY_DATA_PORT = data;
            
            // RB4 can only be an input, so use RB7
            _LATB7 = data >> 4;
            
            latchTimer = GetTimer_uS();    // Capture the timer value
            CodeOutputState = CODE_LATCH_DELAY;
            break;

        case CODE_LATCH_DELAY:

            // Wait latch delay
            if (labs(GetTimer_uS() - latchTimer) >= LATCH_DELAY_uS)
            {
                #if KEYBOARD_STROBE_LEVEL == STROBE_LOW
                STROBE_OUT_SetLow();           // Strobe pulse low
                #elif  KEYBOARD_STROBE_LEVEL == STROBE_HIGH
                STROBE_OUT_SetHigh();          // Set strobe high
                #endif
                latchTimer = GetTimer_uS();     // Capture the timer value
                CodeOutputState = CODE_STROBE_ACTIVE;
            }
            break;

        case CODE_STROBE_ACTIVE:

            // Wait while strobe is active
            if (labs(GetTimer_uS() - latchTimer) >= STROBE_PULSE_WIDTH_uS)
            {
                #if KEYBOARD_STROBE_LEVEL == STROBE_LOW
                STROBE_OUT_SetHigh();          // Strobe - normally high - pulse low
                #elif  KEYBOARD_STROBE_LEVEL == STROBE_HIGH
                STROBE_OUT_SetLow();           // Set strobe low
                #endif
                ParallelPortRateTimer = GetTimer_uS(); // Capture the timer value
                CodeOutputState = CODE_OUT_MAX_RATE_DELAY;
            }
            break;

        case CODE_OUT_MAX_RATE_DELAY:

            // Wait for parallel port rate timer, so we don't send too fast
            if (labs(GetTimer_uS() - ParallelPortRateTimer) >= PARALLEL_MAX_RATE_uS)
            {
                CodeOutputState = CODE_OUT_IDLE;
            }
            break;

        default:

            break;
    }

    /* Repeat key send when active */
    //if ( (keyRepeatActive) && (labs(GetTimer_uS() - keyRepeatTimer) > KEY_REPEAT_RATE_uS) )
    //{
    //   keyRepeatActive = false;
    //   keyRepeatTimer = GetTimer_uS(); 
    //   SendCharToHost(lastChar); /* Send last character again */
    //}
}
/*
 * @brief Send ascii_code 
 * @details Sends ascii_code to parallel port and serial port
 */
void SendCharToHost(uint8_t ch)
{
    #if DEBUG_ASCII_CODE == true
    printf("ASCII 0x%02X\r\n", ch);
    #endif

    lastChar = ch; // Save character for resend
    
    // If ready to send
    if (CodeOutCount < (sizeof(CodeOutBuf) - 1))
    {
        // Buffer data byte send
        CodeOutBuf[CodeOutHead].CodeOut = ch; 
        CodeOutBuf[CodeOutHead].OnlyToParallelPort = PARALLEL_AND_SERIAL;
        CodeOutHead++;
        CodeOutHead &= (CMD_OUT_BUFFER_SIZE - 1);
        CodeOutCount++;
    }
}
/*
 * @brief Send a message to the host parallel port
 *        at the maximum rate
 */
void SendStringToHost(const char *msg, bool OnlyParallelPort)
{
    // Message only to parallel port when OnlyParallelPort set
    
    while (*msg)
    {
            // If ready to send
        if (CodeOutCount < (sizeof(CodeOutBuf) - 1))
        {
            // Buffer message
            CodeOutBuf[CodeOutHead].CodeOut = *msg++; 
            CodeOutBuf[CodeOutHead].OnlyToParallelPort =  OnlyParallelPort;
            CodeOutHead++;
            CodeOutHead &= (CMD_OUT_BUFFER_SIZE - 1);
            CodeOutCount++;
        }
        else
        {
            // Abort if no room in buffer
            break;
        }
    }
}

// EOF